from __future__ import annotations
import math

from multiprocessing import Pool
# from util.mphelper import DummyPool as Pool
from typing import Any, Callable, List, Sequence, Tuple, Union

import numpy as np
import scipy.sparse as sp

import GRT
from GRT import GRTData
from mesh import Domain, Linspace, Mesh
from util import mphelper, stringutils


def _create_shared_info(grt_data: GRTData, y_min: float, s: Sequence[float], n_func: Callable[[np.ndarray, Any], np.ndarray], res_info: Tuple) -> None:
    """create the shared arrays for multiprocessing"""
    return grt_data.dx, mphelper.np_to_shared_info(grt_data.phi), mphelper.np_to_shared_info(grt_data.integrand_factor),\
        y_min, mphelper.array_to_shared(s), n_func, res_info

def __init_worker(dx, phi_info, integrand_factor_info, y_min, s, n_func, res_info):
    # use global variables to save arrays and other infos
    global __grt_data, __y_min, __s, __n_func, __res
    __grt_data = GRTData(dx, mphelper.np_from_shared(phi_info), mphelper.np_from_shared(integrand_factor_info))
    __y_min, __s, __n_func, __res = y_min, s, n_func, mphelper.np_from_shared(res_info)

def __worker_func(t: float, t_i: int, y_i: int, *n_args):
    # write result row into shared array
    GRT.F_tilde(__grt_data, __y_min, __s, t, __n_func, *n_args, res=__res[y_i, :, t_i])

def calculate_kernels_yaxis(grt_data: GRTData, y_min: float,
                        s: Linspace, t: Sequence[float], p: Union[float, Sequence[float]],
                        mollifier: Callable[[np.ndarray, Tuple[float, float]], np.ndarray]) -> Union[np.ndarray, Sequence[np.ndarray]]:
    """
    Calculates the reconstruction kernel for points on the y-axis.

    Parameters
    ----------
        grt_data : GRTData
            return value of 'calculate_F_data'
        y_min : float
            the upper bound of the domain where phi was calculated
        s : linspace sequence of floats
            offsets for the kernels
        t : sequence of floats
            times for which the contours are calculated (for example a suitable 'np.linspace' array)
        p : single float or sequence of floats
            a single y-coordinate or a list of y-coordinates at which the kernels are calculated
        mollifier: function
            K*mollifier - function (see pseudo_ops.py)

    Returns
    ----------
        K : ndarray or list of ndarray
            kernel-array or list of kernel-arrays for each point in 'p'
    """
    # TODO: maybe do the mirroring outside this function
    # extract right half of the s-array
    mid_contained = len(s) % 2 == 1 # uneven array contains the middle
    if not mid_contained:
        print("[kernel.py] Warning: middle of s-domain is not contained in s-array (len(s) is even). This could lead to unwanted results.")
    s_arr = s[len(s) // 2:]

    # p is only a single point
    single = False
    if isinstance(p, float) or isinstance(p, int):
        single = True
        p = [p]

    # prepare result array
    res, res_info = mphelper.create_shared((len(p), len(s_arr), len(t)), float)
    shared_info = _create_shared_info(grt_data, y_min, s_arr, mollifier, res_info)

    with Pool(initializer=__init_worker, initargs=shared_info) as pool:
        print("[kernel.py] Calculating {} kernels using {} cores.".format(len(p), pool._processes))
        bar = stringutils.ProgressBar(60, len(p) * len(t))

        futures = []
        # TODO: maybe remove the first for loop (do everything regarding a single "t" in the worker)
        for p_i, y in enumerate(p):
            for t_i in range(len(t)):
                fut = pool.apply_async(__worker_func, (t[t_i], t_i, p_i, (s.mid, y)))
                futures.append(fut)

        for _ in mphelper.as_completed(futures):
            bar.tick()
        pool.close()
        pool.join()

    #res_sp = []
    #for k in res:
    #    k_sp = sp.csr_matrix(k)
    #    if mid_contained:
    #        res_sp.append(sp.vstack((k_sp[:0:-1], k_sp)))
    #    else:
    #        res_sp.append(sp.vstack((k_sp[::-1], k_sp)))
    #return res_sp[0] if single else res_sp

    # mirror the result into negative 's'-direction
    if mid_contained:
        res = np.concatenate((res[:, :0:-1], res), axis=1)
    else:
        res = np.concatenate((res[:, ::-1], res), axis=1)
    return res[0] if single else res

def calculate_reference_kernels(mesh: Mesh, p_domain: Domain, c: np.ndarray, alpha: float,
                        s: Linspace, t: Sequence[float], y_ref: Sequence[float],
                        mollifier: Callable[[np.ndarray, Tuple[float, float]], np.ndarray], fs_tol: float=1e-3) -> Tuple[List[np.ndarray], Sequence[float], GRTData]:
    """
    calculates the kernels for a sequence of reference y-coordinates y.

    Parameters
    ----------
        mesh : Mesh
            the domain mesh of c.
        p_mesh : Mesh
            the domain mesh for the reconstruction.
        c : ndarray
            the background velocity (needs to be only depth dependent)
        alpha : float
            offset for the common-offset geometry
        s : Linspace
            offsets for the kernel
        t : sequence of floats
            list of times for which the contours are calculated (for example a suitable 'np.linspace' array)
        y_ref : sequence of floats
            a list of reference y-coordinates at which the kernels are calculated
        mollifier: function
            K* mollifier - function
        fs_tol: float
            tolerance for the fast sweeping algorithm (amplitude)

    Returns
    ----------
        kernels : list of NumPy-arrays
            reference reconstruction kernels for every reference y-coordinate.
            shape[0] will be larger than mesh.n[0] because of slicing in 'reconstruction.scalar_product'.
        s_ref : Linspace
            offsets of the reference kernels (will contain s)
        grt_data : GRTData
            data for evaluating F*n
    """
    assert mesh.n == c.shape, "mesh and velocity do not have the same shape"
    assert mesh.domain[0, 0] == 0, "x-start of 'mesh' must be 0"
    assert np.all(p_domain.start >= (s[0], y_ref[0])), "'p_mesh' must be contained in the 's-y_ref mesh'"
    assert np.all(p_domain.end <= (s[-1], y_ref[-1])), "'p_mesh' must be contained in the 's-y_ref mesh'"

    # TODO: maybe function for c as argument
    n_additional = math.ceil(alpha / mesh.dx[0])
    # if alpha > mesh.width, c needs to be more than doubled
    c_stack = [c]
    while n_additional > c.shape[0]:
        c_stack.append(c)
        n_additional -= c.shape[0]
    c_stack.append(c[:n_additional])
    # stack c along the s-axis
    c_ref = np.vstack(c_stack)

    grt_data = GRT.calculate_F_data(c_ref, mesh.dx, alpha, fs_tol=fs_tol)

    # we need to expand 's' by the maximum axis of the p-mesh
    max_p_axis = max(s.mid - p_domain[0, 0], p_domain[0, 1] - s.mid)
    # try to account for rounding errors later on
    if (max_p_axis / s.dx) % 1 != 0:
        max_p_axis += s.dx
    s_ref = s.expanded(max_p_axis, max_p_axis)

    print(f"""Calculating reference kernel(s) on s-t mesh
    [{s_ref[0]}, {s_ref[-1]}] x [{t[0]}, {t[-1]}]
    dx = ({s.dx}, {t[1]-t[0]})
    n = {len(s_ref)} x {len(t)}""")
    # TODO: maybe don't use mesh as y_min?
    kernels_ref = calculate_kernels_yaxis(grt_data, mesh.domain[1, 0], s_ref, t, y_ref, mollifier)

    return kernels_ref, s_ref, grt_data
