# Factored Amplitude

amplitude is a Python (C++-)extension which implements the Lax-Friedrich sweeping method
and is used to solve the transport equation
    
    2 grad(a) grad(tau) + a laplace(tau) = 0.

a and tau are factored into `a = a0 a1` and `tau = tau0 tau1`,
where

    a0(x) = 1 / (2 pi sqrt(|x - x_s|),
    tau0(x) = |x - x_s|,

and `tau1` is the solution of the factored eikonal equation.

## References
- Luo, S., Qian, J., and Zhao, H.. *Higher-order schemes for 3-D traveltimes and amplitudes*. Geophysics, 77. T47–T56 (2012). https://doi.org/10.1190/geo2010-0363.1


# Requirements

- Python 3
- numpy version 1.7 or higher
- C++11 compiler


# Installation
  
```bash
pip install .
```


# Examples

```python
import numpy as np
import amplitude

c = np.ones((100, 100))
tau1 = np.ones((100, 100))
x_s = (0, 0)
dx = (1.0, 1.0)
order = 2

a1 = amplitude.factored_amplitude(c, tau1, x_s, dx, 1e-5)
```

Note that the source position `x_s` describes an index-vector.

To visualize the results, matplotlib (https://pypi.org/project/matplotlib/) can be used, for example:

```python
import matplotlib.pyplot as plt

# for the base amplitude function 'x_s' describes real coordinates
a0 = amplitude.base_amplitude(a1.shape, dx, x_s, indexing="ij")
plt.contourf(a0 * a1)
plt.show()
```
