"""
Factored Amplitude
--------

Amplitude is a Python (C++-)extension which implements the Lax-Friedrich sweeping method
and is used to solve the transport equation
    2 grad(a) grad(tau) + a laplace(tau) = 0.
`a` and `tau` are factored into `a = a0 a1` and `tau = tau0 tau1`,
where
    a0(x) = 1 / (2 sqrt(2 pi) sqrt(|x - x_s|),
    tau0(x) = |x - x_s|,
and `tau1` is the solution of the factored eikonal equation.
"""
__version__ = "0.2.0"

import numpy as np
from .camplitude import factored_amplitude
from typing import Sequence


def base_amplitude(shape: Sequence[int], dx: Sequence[float], x_s: Sequence[int], indexing: str="xy") -> np.ndarray:
    """
    Calculates the base amplitude `a0(x) = 1 / (2 sqrt(2 pi) sqrt(|x - x_s|)` according to the euclidian metric.

    Parameters
    ----------
        shape : sequence of ints
            Shape of the generated array.
        dx : sequence of doubles
            Grid spacing for each dimension, dx > 0.
            Must have the same length as the number of dimensions of shape.
        x_s : sequence of ints
            Source position as position-vector, e.g. `(0.5, 1)`.
            Must have the same length as the number of dimensions of shape.
        indexing : {'xy', 'ij'}, optional
            Cartesian ('xy', default) or matrix ('ij') indexing of output.
            See numpy's meshgrid function for more details.

    Returns
    ----------
        a0 : ndarray
            Base amplitude `1 / (2 sqrt(2 pi) sqrt(|x - x_s|)`.
    """
    assert len(shape) == len(dx) == len(x_s)
    x = []
    for shape_i, dx_i in zip(shape, dx):
        x.append(np.linspace(0, shape_i*dx_i, shape_i, endpoint=False))

    mesh = np.array(np.meshgrid(*x, indexing=indexing))
    dx = np.asarray(dx)
    distance = np.linalg.norm((mesh.T - x_s * dx).T, ord=2, axis=0)
    
    # avoid division by 0
    distance[tuple(x_s)] = (dx[0] + dx[1]) / 8
    return 1 / (2 * np.sqrt(2*np.pi) * np.sqrt(distance))
