#include <cmath>
#include "hamiltonians.hpp"
#include "utils.hpp"

/**
 * calculates the laplacian of 'f' in a point 'p' using second order central finite differences on the inside
 * and first order one sided differences on the boundary.
 * @param f the function/array to calculate the laplacian for
 * @param p the position
 * @param p_i the flattened index position
 * @param shape the shape of the mesh
 * @param dx the axis-grid spacing
 * @param shift the shift to apply in each axis to get neighboring indices
 * @return finite difference approximation Delta f(z)
 */
template <int NDIM>
inline auto laplacian(const double* const f, const Point<NDIM> p, const usize p_i, const usize (&shape)[NDIM], const double (&dx)[NDIM], const ssize *shift) -> double
{
    double res = 0;
    for (int d=0; d < NDIM; d++)
    {
        res += lapl(f, p[d], p_i, shape[d], dx[d], shift[d]);
    }
    return res;
}


#define EPS 0.001

template<>
void FactoredAmplitude<2>::initialize(const SweepInfo<2> &info, const usize x0, double *const res, std::array<std::vector<double>, 2> &alpha)
{
    const double SQRT_PI = sqrt(2 * M_PI);
    auto p_x0 = info.get_vector(x0);

    for (usize x=0; x < info.shape[0]; x++)
    {
        for (usize y=0; y < info.shape[1]; y++)
        {
            Point<2> p_x{x, y};
            usize x_i = info.get_index(p_x);
            
            // initialize result array
            res[x_i] = sqrt(c[x_i]);
            
            // we don't need Hdu, Hdp and Hdq-values near the source (the values are fixed anyways)
            if ((labs(x - p_x0[0]) <= 1) && (labs(y - p_x0[1]) <= 1))
            {
                alpha[0][x_i] = EPS;
                alpha[1][x_i] = EPS;
                continue;
            }
            
            double X = (ssize)(x - p_x0[0]) * info.dx[0];
            double Y = (ssize)(y - p_x0[1]) * info.dx[1];
            double tau0 = pow(pow(X, 2) + pow(Y, 2), 0.5);
            double tau1_dx = diff(tau_1, x, x_i, info.shape[0], info.dx[0], info.shift[0]);
            double tau1_dy = diff(tau_1, y, x_i, info.shape[1], info.dx[1], info.shift[1]);

            Hdu[x_i] = pow(tau0, -1.5) * (tau1_dx * X + tau1_dy * Y) / (2*SQRT_PI) +\
                pow(tau0, 0.5) * laplacian<2>(tau_1, p_x, x_i, info.shape, info.dx, info.shift) / (2*SQRT_PI);

            Hdp[0][x_i] = pow(tau0, 0.5) * tau1_dx / SQRT_PI + pow(tau0, -1.5) * X * tau_1[x_i] / SQRT_PI;
            Hdp[1][x_i] = pow(tau0, 0.5) * tau1_dy / SQRT_PI + pow(tau0, -1.5) * Y * tau_1[x_i] / SQRT_PI;

            // update alpha values (+offset for numerical inaccuracy)
            // LLF (local lax friedrichs)
            alpha[0][x_i] = EPS + fabs(Hdu[x_i]) + fabs(Hdp[0][x_i]);
            alpha[1][x_i] = EPS + fabs(Hdu[x_i]) + fabs(Hdp[1][x_i]);
        }
    }

    delete[] p_x0;

    // SLF (stencil lax friedrichs)
    // int k = 3;
    // double a;
    // for (usize i=0; i < info.size; i++)
    // {
    //     double max_x = 0, max_y = 0;
    //     Point<2> x_p = info.get_point(i);
    //
    //     for (int x_off=-(int)std::min<usize>(k, x_p[0]); x_off <= (int)(info.shape[0] - std::max<usize>(info.shape[0]-1, x_p[0]+k)); x_off++)
    //     {
    //         for (int y_off=-(int)std::min<usize>(k, x_p[1]); y_off <= (int)(info.shape[1] - std::max<usize>(info.shape[1]-1, x_p[1]+k)); y_off++)
    //         {
    //             usize j = i + info.get_index<int>({x_off, y_off});
    //             a = 0.01 + fabs(Hdu[j]) + fabs(Hdp[0][j]);
    //             if (a > max_x)
    //                 max_x = a;
    //             a = 0.01 + fabs(Hdu[j]) + fabs(Hdp[1][j]);
    //             if (a > max_y)
    //                 max_y = a;
    //         }
    //     }
    //
    //     alpha[0][i] = max_x;
    //     alpha[1][i] = max_y;
    // }

    // GLF (global lax friedrichs)
    double max_x = 0, max_y = 0;
    for (usize i=0; i < info.size; i++)
    {
        if (alpha[0][i] > max_x)
        {
            max_x = alpha[0][i];
        }
        if (alpha[1][i] > max_y)
        {
            max_y = alpha[1][i];
        }
    }
    for (usize i=0; i < info.size; i++)
    {
        alpha[0][i] = max_x;
        alpha[1][i] = max_y;
    }
}

template<>
void FactoredAmplitude<3>::initialize(const SweepInfo<3> &info, const usize x0, double *const res, std::array<std::vector<double>, 3> &alpha)
{
    const double SQRT_PI = sqrt(2 * M_PI);
    auto p_x0 = info.get_vector(x0);

    for (usize x=0; x < info.shape[0]; x++)
    {
        double X = (ssize)(x - p_x0[0]) * info.dx[0];

        for (usize y=0; y < info.shape[1]; y++)
        {
            double Y = (ssize)(y - p_x0[1]) * info.dx[1];

            for (usize z=0; z < info.shape[2]; z++)
            {
                Point<3> p_x{x, y, z};
                usize x_i = info.get_index(p_x);
                
                // initialize result array
                res[x_i] = sqrt(c[x_i]);
                
                // we don't need Hdu, Hdp and Hdq-values near the source (the values are fixed anyways)
                if ((labs(x - p_x0[0]) <= 1) && (labs(y - p_x0[1]) <= 1) && (labs(z - p_x0[2]) <= 1))
                {
                    alpha[0][x_i] = EPS;
                    alpha[1][x_i] = EPS;
                    alpha[2][x_i] = EPS;
                    continue;
                }

                double Z = (ssize)(z - p_x0[2]) * info.dx[2];
                double tau0 = pow(pow(X, 2) + pow(Y, 2) + pow(Z, 2), 0.5);
                double tau1_dx = diff(tau_1, x, x_i, info.shape[0], info.dx[0], info.shift[0]);
                double tau1_dy = diff(tau_1, y, x_i, info.shape[1], info.dx[1], info.shift[1]);
                double tau1_dz = diff(tau_1, z, x_i, info.shape[2], info.dx[2], info.shift[2]);

                Hdu[x_i] = pow(tau0, -1.5) * (tau1_dx * X + tau1_dy * Y + tau1_dz * Z) / (2*SQRT_PI) +\
                    pow(tau0, 0.5) * laplacian<3>(tau_1, p_x, x_i, info.shape, info.dx, info.shift) / (2*SQRT_PI);

                Hdp[0][x_i] = pow(tau0, 0.5) * tau1_dx / SQRT_PI + pow(tau0, -1.5) * X * tau_1[x_i] / SQRT_PI;
                Hdp[1][x_i] = pow(tau0, 0.5) * tau1_dy / SQRT_PI + pow(tau0, -1.5) * Y * tau_1[x_i] / SQRT_PI;
                Hdp[2][x_i] = pow(tau0, 0.5) * tau1_dz / SQRT_PI + pow(tau0, -1.5) * Z * tau_1[x_i] / SQRT_PI;

                // update alpha values (+offset for numerical inaccuracy)
                // LLF (local lax friedrichs)
                alpha[0][x_i] = EPS + fabs(Hdu[x_i]) + fabs(Hdp[0][x_i]);
                alpha[1][x_i] = EPS + fabs(Hdu[x_i]) + fabs(Hdp[1][x_i]);
                alpha[2][x_i] = EPS + fabs(Hdu[x_i]) + fabs(Hdp[2][x_i]);
            }
        }
    }

    delete[] p_x0;

    // GLF (global lax friedrichs)
    double max_x = 0, max_y = 0, max_z = 0;
    for (usize i=0; i < info.size; i++)
    {
        if (alpha[0][i] > max_x)
        {
            max_x = alpha[0][i];
        }
        if (alpha[1][i] > max_y)
        {
            max_y = alpha[1][i];
        }
        if (alpha[2][i] > max_z)
        {
            max_z = alpha[2][i];
        }
    }
    for (usize i=0; i < info.size; i++)
    {
        alpha[0][i] = max_x;
        alpha[1][i] = max_y;
        alpha[2][i] = max_z;
    }
}

template <int NDIM>
auto FactoredAmplitude<NDIM>::H(const usize x, const double u, const double (&du)[NDIM]) const -> double
{
    double res = u * Hdu[x];
    for (int d=0; d < NDIM; d++)
    {
        res += du[d] * Hdp[d][x];
    }
    return res;
}

// explicitly instanciate template function
template auto FactoredAmplitude<2>::H(const usize x, const double u, const double (&du)[2]) const -> double;
template auto FactoredAmplitude<3>::H(const usize x, const double u, const double (&du)[3]) const -> double;


// template <>
// void FactoredEikonal<2>::initialize(const SweepInfo<2> &info, const usize x0, double *const res, std::array<std::vector<double>, 2> &alpha)
// {
//     auto p_x0 = info.get_vector(x0);

//     for (usize x=0; x < info.shape[0]; x++)
//     {
//         for (usize y = 0; y < info.shape[1]; y++)
//         {
//             Point<2> p_x{x, y};
//             usize x_i = info.get_index(p_x);

//             X[0][x_i] = (ssize)(x - p_x0[0]) * info.dx[0];
//             X[1][x_i] = (ssize)(y - p_x0[1]) * info.dx[1];

//             // initialize result array
//             // if ((labs((size)(p[0] - x0[0])) <= 1) and (labs((size)(p[1] - x0[1])) <= 1))
//                 // TODO: directional dx
//             res[x_i] = 1 / c[x_i];
//             // else
//             //     res[p.z] = 1;

//             alpha[0][x_i] = c[x_i] / 1; // s / s_0 (see email qian)
//             alpha[1][x_i] = c[x_i] / 1;
//         }
//     }
//     delete[] p_x0;

//     // GLF (global lax friedrichs)
//     double max_x = 0, max_y = 0;
//     for (usize i=0; i < info.size; i++)
//     {
//         if (alpha[0][i] > max_x)
//         {
//             max_x = alpha[0][i];
//         }
//         if (alpha[1][i] > max_y)
//         {
//             max_y = alpha[1][i];
//         }
//     }
//     for (usize i=0; i < info.size; i++)
//     {
//         alpha[0][i] = max_x;
//         alpha[1][i] = max_y;
//     }
// }

// template <>
// void FactoredEikonal<3>::initialize(const SweepInfo<3> &info, const usize x0, double *const res, std::array<std::vector<double>, 3> &alpha)
// {
//     // NO OP
// }

// template <int NDIM>
// auto FactoredEikonal<NDIM>::H(const usize x, const double u, const double (&du)[NDIM]) const -> double
// {
//     double tau0_sq = 0;
//     double du_norm_sq = 0;
//     double sprod = 0;

//     for (int d=0; d<NDIM; d++)
//     {
//         tau0_sq += pow(X[d][x], 2);
//         du_norm_sq += pow(du[d], 2);
//         sprod += X[d][x] * du[d];
//     }

//     double exp = tau0_sq * du_norm_sq + 2*u * sprod + pow(u, 2);

//     return pow(exp, 0.5) - 1/c[x];
// }
// // explicitly instanciate template function
// template auto FactoredEikonal<2>::H(const usize x, const double u, const double (&du)[2]) const -> double;
// template auto FactoredEikonal<3>::H(const usize x, const double u, const double (&du)[3]) const -> double;