#include "typedefs.hpp"


/**
 * calculates the derivative of 'f' in a single axis using second order central finite differences on the inside
 * and first order one sided differences on the boundary.
 * @param f the function/array to differentiate
 * @param x_i the position in the axis
 * @param x the flattened index position
 * @param shape the shape of the axis
 * @param dx the axis-grid spacing
 * @param shift the shift to apply to 'x' to get to the neighbors in this axis
 * @return finite difference approximation (f'(z))_axis
 */
inline auto diff(const double* const f, const usize x_i, const usize x, const usize shape, const double dx, const usize shift) -> double
{
    if (x_i == 0)
    {
        return (f[x + shift] - f[x]) / dx;
    }
    else if (x_i == shape-1)
    {
        return (f[x] - f[x - shift]) / dx;
    }
    else
    {
        return (f[x + shift] - f[x - shift]) / (2 * dx);
    }
}

/**
 * calculates the second derivative of 'f' in a single axis using second order central finite differences on the inside
 * and first order one sided differences on the boundary.
 * @param f the function/array
 * @param x_i the position in the axis
 * @param x the flattened index position
 * @param shape the shape of the axis
 * @param dx the axis-grid spacing
 * @param shift the shift to apply to 'z' to get to the neighbors in this axis
 * @return finite difference approximation (f''(z))_axis
 */
inline auto lapl(const double* const f, const usize x_i, const usize x, const usize shape, const double dx, const usize shift) -> double
{
    if (x_i == 0)
    {
        return (f[x] - 2 * f[x + shift] + f[x + 2*shift]) / pow(dx, 2);
    }
    else if (x_i == shape-1)
    {
        return (f[x] - 2 * f[x - shift] + f[x - 2*shift]) / pow(dx, 2);
    }
    else
    {
        return (f[x + shift] - 2 * f[x] + f[x - shift]) / pow(dx, 2);
    }
}

// debug
// void print_data(const SweepInfo<2> &info, double *data, int precision)
// {
//     std::cout << std::endl;
//     std::string fmtstring = "%." + std::to_string(precision) + "f";

//     usize i = 0;
//     for (usize x = 0; x < info.shape[0]; x++)
//     {
//         for (usize y = 0; y < info.shape[1]; y++)
//         {
//             char str[256];
//             sprintf(str, fmtstring.c_str(), data[i++]);
//             std::cout << str << "\t";
//         }
//         std::cout << std::endl;
//     }
// }
