from configparser import ConfigParser
import re
from typing import Callable, Optional
from mesh import Domain, Mesh, Linspace
import rhs
import numpy as np


# needed on systems without 'fork'
class _PickleableFunction:
    def __init__(self, args: str, body: str, name: str="pickleable_func"):
        self.name = name
        self.func_str = f"def {name}({args}):\n    " + body.replace("\n", "\n    ")
        self.func = self._exec()

    def _exec(self):
        context = {}
        exec(self.func_str, {'np': np, 'rhs': rhs}, context)
        return context[self.name]
    
    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __getstate__(self):
        return {k: v for k, v in self.__dict__.items() if k != "func"}
    
    def __setstate__(self, state):
        self.__dict__ = state
        self.func = self._exec()

    def __str__(self):
        return self.func_str


class Config(ConfigParser):
    def __init__(self, path, **kwargs):
        super(ConfigParser, self).__init__(**kwargs)
        read_files = self.read(path)
        if path not in read_files:
            raise Exception(f"Could not read config file '{path}'")

    def geteval(self, section: str, option: str, *args, raw: bool=False, vars: Optional[dict]=None, namespace={}):
        value = self.get(section, option, *args, raw=raw, vars=vars)
        namespace.update({'np': np, 'Domain': Domain, 'Mesh': Mesh, 'Linspace': Linspace})
        return eval(value, namespace)
    
    def getlambda(self, section: str, option: str, *f_args: str) -> Callable:
        f_str = f"lambda " + ", ".join(f_args) + ": " + self.get(section, option)
        return eval(f_str, {'np': np})
    
    def getvelocity(self, section: str, option: str, mesh: Mesh):
        v = np.zeros(mesh.n)

        def Ball(m, r, value, ord=2):
            rhs.add_ball(v.T, mesh.X.T, m, r, value, ord=ord)

        def Sin(y, a, p, o, value):
            flag = mesh.X[1] > y + a * np.sin(p * mesh.X[0] + o)
            v[flag] += value
        
        def Floor(y, value):
            v[mesh.X[1] > y] += value

        def Rect(m, r, value):
            rhs.add_rect(v.T, mesh.X.T, m, r, value)
        
        exec(self.get(section, option), {'np': np, 'Ball': Ball, 'Sin': Sin, 'Floor': Floor, 'Rect': Rect})
        return v
    
    def getnfunc(self, section: str, option: str):
        n_str = self.get(section, option)
        n_str = re.sub(r"Ball\((.*)\)", r"rhs.add_ball(n, x, \1)", n_str)
        n_str = re.sub(r"Sin\((.*)\)", r"rhs.add_sin(n, x, \1)", n_str)
        n_str = re.sub(r"Floor\((.*)\)", r"rhs.add_floor(n, x, \1)", n_str)
        n_str = re.sub(r"Rect\((.*)\)", r"rhs.add_rect(n, x, \1)", n_str)

        n_func = _PickleableFunction("x", "n = np.zeros(len(x))\n" + n_str + "\nreturn n""", "phantom_func")
        return n_func
    
    def __str__(self):
        _indent = "   "
        res = "Config:"
        for sec, entries in self.items():
            res += f"\n[{sec}]"
            for k, v in entries.items():
                v = v.replace("\n", "\n" + " "*(3 + len(_indent) + len(k)))
                res += f"\n{_indent}{k} = {v}"
            res += "\n"
        return res