from __future__ import annotations
import math
import decimal
from typing import Any, Generic, Iterable, Iterator, List, Optional, Tuple, TypeVar


class ToleranceDict:
    def __init__(self, base: float) -> None:
        self._container = {}
        self._base = base
    
    def _convert_key(self, key: float) -> float:
        # converts the given key to it's representation as multiple of self._tol
        return int(key / self._base + math.copysign(0.5, key))

    def __getitem__(self, key: float) -> Any:
        key = self._convert_key(key)
        return self._container[key]

    def __setitem__(self, key: float, value: Any) -> None:
        key = self._convert_key(key)
        self._container[key] = value
    
    def __contains__(self, key: float) -> bool:
        key = self._convert_key(key)
        return key in self._container
    
    def __str__(self) -> str:
        d = decimal.Decimal(self._base)
        d.normalize()
        d = d.as_tuple()
        decimals = max(-(len(d.digits)-1 + d.exponent), 0)
        fmt_str = r"{:." + str(decimals) + r"f}: {}"

        res = []
        for k, v in self._container.items():
            res.append(fmt_str.format(k * self._base, v))
        
        return "{" + ", ".join(res) + "}"

_K = TypeVar("_K", float, int)
_V = TypeVar("_V")

class InterpolationDict(Generic[_K, _V]):
    def __init__(self, items: Optional[Iterable[Tuple[_K, _V]]]=None, assume_sorted: bool=False) -> None:
        self._keys: List[_K] = []
        self._values: List[_V] = []

        if items is not None:
            if not assume_sorted:
                items = sorted(items, key=lambda x: x[0])
            for k, v in items:
                self._keys.append(k)
                self._values.append(v)
    
    def _binary_search(self, key: _K) -> int:
        """ returns:
            -1, if key is smaller then the smallest current key
            len(self._keys) - 1, if key is bigger than or equal to the biggest current key
            l, where for the left index l: keys[l] < key < keys[r]
        """
        if len(self._keys) == 0 or key < self._keys[0]:
            return -1
        elif key > self._keys[-1]:
            return len(self._keys) - 1

        l, r = 0, len(self._keys)
        while l + 1 < r:
            m = l + (r - l) // 2
            if key == self._keys[m]:
                return m
            elif key < self._keys[m]:
                r = m
            else:
                l = m
        return l

    def __getitem__(self, key: _K) -> _V:
        i = self._binary_search(key)

        if i == -1:
            return self._values[0]
        elif i == len(self._keys) - 1:
            return self._values[-1]
        else:
            gamma = (self._keys[i+1] - key) / (self._keys[i+1] - self._keys[i])
            return gamma * self._values[i] + (1 - gamma) * self._values[i+1]
    
    def __setitem__(self, key: _K, value: _V) -> None:
        i = self._binary_search(key)
        self._keys.insert(i + 1, key)
        self._values.insert(i + 1, value)
    
    def __delitem__(self, key: _K) -> None:
        i = self._binary_search(key)
        if self._keys[i] != key:
            raise KeyError

        del self._keys[i]
        del self._values[i]
    
    def keys(self) -> Iterator[_K]:
        yield from self._keys
    
    def values(self) -> Iterator[_V]:
        yield from self._values
    
    def items(self) -> Iterator[Tuple[_K, _V]]:
        return zip(self._keys, self._values)
    
    def __iter__(self) -> Iterator[_K]:
        return self.keys()
    
    def __str__(self) -> str:
        return str(dict(zip(self._keys, self._values)))


if __name__ == "__main__":
    # s = ToleranceDict(10)

    # s[1] = "Eins"
    # s[2] = "Zwei"
    # s[2.5] = 2.5
    # s[2.55] = 2.55
    # s[2.555] = 2.555
    # s[1e-5] = 1e-5
    # s[2e-5] = 2e-5

    # print(s)
    # print()


    ipd = InterpolationDict()
    ipd[0] = 10
    ipd[1] = 20
    ipd[2] = 10

    import numpy as np
    p_ref, _ = np.linspace(0, 7, num=51, retstep=True)
    F = range(len(p_ref))
    F_p = InterpolationDict(zip(p_ref, F))

    for y in p_ref:
        print(F_p[y])