from typing import Optional
import pickle
import os
import re
import numpy as np


# current working directory is the base path
base_path = os.path.abspath("") + os.path.sep

def res_path(*folder: str) -> str:
    """Returns a resource path and creates it if it's not yet present."""
    path = base_path
    if len(path) > 0:
        path = os.path.normpath(os.path.join(path, *folder))
    path += os.path.sep

    if not os.path.exists(path):
        os.makedirs(path)
    return path

def _next_filename(name, ftype, fnames):
    endings = set()
    # filter files
    for fn in fnames:
        match = re.fullmatch(name + r"([0-9]*)\." + ftype, fn)
        if match:
            endings.add(match.group(1))

    # find next available file name
    i = 1
    while str(i) in endings:
        i += 1
    return name + str(i) +"."+ftype

def next_file(path, name):
    fname, ftype = name.rsplit(".", 1)
    return os.path.join(path, _next_filename(fname, ftype, os.listdir(path)))

# -------------------------------------------------
# Plot saving
# -------------------------------------------------
base_path_plots = os.path.join(base_path, "figures") + os.path.sep

def fig_path(name: Optional[str]=None) -> str:
    """Returns a figure path for the folder name 'name' and creates it if it's not yet present."""
    return res_path("figures", name)

# -------------------------------------------------
# Array saving/loading
# -------------------------------------------------
_file_type = ".pcl"

def load_scenario(identifier, sim_func, match_func):
    base_path = res_path("saved")
    files = [fn for fn in os.listdir(base_path) if re.fullmatch(identifier + r"[0-9]*", fn)]
    
    # check if scenario was saved before
    for fn in files:
        with open(os.path.join(base_path, fn), "rb") as f:
            data = pickle.load(f)
            if match_func(data):
                print("loaded data from " + os.path.join(base_path, fn))
                return data
    
    print("no data found. Calculating...")
    data = sim_func()
    print("Finished calculating...")
    
    fn = _next_filename(identifier, _file_type, files)
    
    with open(os.path.join(base_path, fn), "wb") as f:
        print("saving data to " + os.path.join(base_path, fn))
        pickle.dump(data, f)
        
    return data

def save_data(identifier, data, ending=".pcl"):
    subf, name = os.path.split(identifier)
    path = res_path(os.path.join("saved", subf)) + name+ending
    print("saving data to " + path)
    with open(path, "wb") as f:
        pickle.dump(data, f)

def load_data(identifier, ending=".pcl"):
    subf, name = os.path.split(identifier)
    path = res_path(os.path.join("saved", subf)) + name+ending
    print("loading data from " + path)
    with open(path, "rb") as f:
        return pickle.load(f)

def save_array(identifier, array):
    subf, name = os.path.split(identifier)
    path = res_path(os.path.join("saved", subf)) + name+".gz"
    print("saving array to " + path)
    np.savetxt(path, array)

def load_array(identifier):
    subf, name = os.path.split(identifier)
    path = res_path(os.path.join("saved", subf)) + name+".gz"
    print("loading array from " + path)
    return np.loadtxt(path)
