from typing import Optional
import matplotlib.pyplot as plt
import os
import subprocess


def _name_gen_func():
    x = 0
    while True:
        yield "figure" + str(x)
        x += 1
_name_gen = _name_gen_func()


class Pdf:
    def __init__(self, documentclass="a4paper", language="english", output_path="output"):
        self._output_dir = os.path.abspath(os.path.dirname(output_path))
        self._output_name = os.path.splitext(os.path.basename(output_path))[0] or "output"

        self._header = r"\documentclass[12pt, " + documentclass + r""", parskip]{scrartcl}
\usepackage[utf8]{inputenc}
\usepackage[T1]{fontenc}
\usepackage[""" + language + r"""]{babel}
\usepackage{float}
\usepackage[pdftex]{graphicx}
\usepackage{amsmath, amssymb, amsthm}
\usepackage{listings}
\usepackage{xcolor}

\lstset{%
    backgroundcolor=\color{gray!10},
    basicstyle=\footnotesize,
    tabsize=4,
    columns=fullflexible,
    frame=none,
    keepspaces=true
}

\begin{document}"""

        self._content = ""
        self._footer = r"\end{document}"

        self._figures = {}
    
    def add_text(self, text: str) -> None:
        if self._content:
            self._content += "\n\n"
        self._content += text
    
    def add_images(self, *figures: plt.Figure, caption: Optional[str]=None, label: Optional[str]=None, width: Optional[int]=1.0) -> None:
        assert len(figures) > 0

        if self._content:
            self._content += "\n"
        self._content += r"""\begin{figure}[H]
    \centering
"""
        scale = width / len(figures)
        for figure in figures:
            name = self._output_name +"_"+ next(_name_gen)
            self._figures[name] = figure
            self._content += "\t" + r"\includegraphics[width=" + f"{scale:.2f}" + r"\textwidth]{" + name + ".pdf}\n"

        if caption:
            self._content += "\t" + r"\caption{" + caption + "}\n"
        if label:
            self._content += "\t" + r"\label{" + label + "}\n"
        self._content += r"\end{figure}"
    
    def add_listing(self, text: str) -> None:
        if self._content:
            self._content += "\n"
        self._content += r"""\begin{lstlisting}
""" + text + r"""
\end{lstlisting}"""

    def compile(self, cleanup=True, del_resources=False) -> None:
        output_path = os.path.join(self._output_dir, self._output_name)

        if not os.path.exists(self._output_dir):
            os.makedirs(self._output_dir)

        # save figures as temporary image files
        for f_name, fig in self._figures.items():
            fig.savefig(os.path.join(self._output_dir, f_name+".pdf"), orientation="landscape", bbox_inches="tight", pad_inches=0.01)
            plt.close(fig)

        with open(output_path + ".tex", "w") as f:
            f.write(str(self))

        try:
            subprocess.run(["pdflatex", self._output_name], cwd=self._output_dir, check=True)
        finally:
            if cleanup:                
                # pdflatex log files
                os.remove(output_path + ".log")
                os.remove(output_path + ".aux")

            if del_resources:
                # remove image files
                for f_name, fig in self._figures.items():
                    os.remove(os.path.join(self._output_dir, f_name+".pdf"))
                os.remove(output_path + ".tex")
        
        print(f"[TeXgen] pdf saved to {output_path}.pdf")

    def __str__(self):
        return self._header + "\n" + self._content + "\n" + self._footer


def mesh_to_tex(mesh: "Mesh") -> str:
    return str(mesh).replace("    ", "").replace("\n", "\\\\\n\indent\hspace{1cm}")


if __name__ == "__main__":
    import argparse
    parser = argparse.ArgumentParser(usage="%(prog)s [FILE]")
    parser.add_argument("file")
    args = parser.parse_args()

    assert args.file.endswith(".tex"), "file must be a .tex file."
    path, fn = os.path.split(args.file)
    name, ext = os.path.splitext(fn)
    
    cleanup = True

    try:
        subprocess.run(["pdflatex", "-halt-on-error", fn], cwd=path, check=True)
    finally:
        path_no_ext = os.path.join(path, name)
        if cleanup:                
            # pdflatex log files
            os.remove(path_no_ext + ".log")
            os.remove(path_no_ext + ".aux")

    print(f"[TeXgen] pdf saved to {path_no_ext}.pdf")